import os

from .utils import Decryption, ByteBuffer
from .utils import ImageAsset, AudioAsset
from .utils import ActionNode, StageNode, ControlSettings, Transition
from .utils import StoryMetadata

class Story:
    def __init__(self, path, uuid, keys:tuple):
        self.mountpoint = path

        self.full_uuid = uuid
        self.uuid = uuid.hex[-8:].upper()

        self.__device_key = keys[0]
        self.__lunii_generic_key = keys[1]

        self.version, self.night_mode, self.authorized = None, False, False
        self.__si, self.__ri, self.__li = None, None, None     

        self.__ri_path = self.__get_local_filepath('ri')
        self.__si_path = self.__get_local_filepath('si')
        self.__li_path = self.__get_local_filepath('li')
        self.__bt_path = self.__get_local_filepath('bt')
        self.__ni_path = self.__get_local_filepath('ni')
        self.__nm_path = self.__get_local_filepath('nm')

        self.__rf_path = self.__get_local_filepath('rf/')
        self.__sf_path = self.__get_local_filepath('sf/')

        self.nodes_index_list = []
        self.stage_nodes = {}

        self.cover_audio = None

        self.populate()

    def populate(self):
        self.__verify_auth()
        self.__get_story_metadata()
        self.__get_story_title()
        self.__get_songs_index()
        self.__get_resources_index()

        self.__get_list_node_index()
        self.__get_action_node_index()

        self.__get_cover_audio()

    def __get_cover_audio(self):
        # reference to AudioAsset of node #0
        self.cover_audio = self.stage_nodes.get(0,{}).assets.get('audio', None)

    def __get_story_metadata(self):
        with open(self.__ni_path, 'rb') as fp_ni:
            ni = fp_ni.read(512)
            buffer = ByteBuffer(ni, '<')
            s1 = buffer.get_short()  # read the first short integer
            s2 = buffer.get_short()  # read the second short integer
            self.version = s2

        # night mode is enabled if the file exists
        self.night_mode = os.path.exists(self.__nm_path)

    def __get_story_title(self):
        storyMetadataClient = StoryMetadata("./misc/titles.json")
        self.title = storyMetadataClient.get_metadata(self.full_uuid).get('title', 'Unknown Story')

    def __get_local_filepath(self, file):
        return self.mountpoint + '/.content/' + self.uuid + '/' + file

    def __get_songs_index(self):
        # Song Index
        # The file is organized as a list of 12 Bytes strings
        # first decrypt the file with generic key
        with Decryption(self.__lunii_generic_key, path=self.__si_path).decrypted_buffer as fp_si:
            loop_again = True
            str_list = []
            while loop_again:
                char_array = fp_si.read(12)
                if char_array:
                    str_list.append(char_array.decode('utf-8'))
                else:
                    loop_again = False
            self.__si = str_list

    def __get_resources_index(self):
        # Resource Index
        # The file is organized as a list of 12 Bytes strings
        # first decrypt the file with generic key
        with Decryption(self.__lunii_generic_key, path=self.__ri_path).decrypted_buffer as fp_ri:
            loop_again = True
            str_list = []
            while loop_again:
                char_array = fp_ri.read(12)
                if char_array:
                    str_list.append(char_array.decode('utf-8'))
                else:
                    loop_again = False
            self.__ri = str_list

    def __get_abs_node_idx(self, node_idx):
        if node_idx < 0:
            return -1
        else:
            return self.nodes_index_list[node_idx]
    

    def __get_list_node_index(self):
        # List Node Index
        # The file is organized as an array of 4 bytes integers
        # first decrypt the file with generic key
        fp_li = Decryption(self.__lunii_generic_key, path=self.__li_path).decrypted_buffer
        bb = ByteBuffer(fp_li.read(), '<')

        #count = fp_li_size_bytes / 4
        count = int(os.path.getsize(self.__li_path) / 4)
        
        self.nodes_index_list = []
        for i in range(count):
            self.nodes_index_list.append(bb.get_int())   


    def __get_action_node_index(self):
        self.stage_nodes = {}
        action_nodes_options_count = {}
        transitions_with_action = {}

        # use ByteBuffer to read the file
        # open ni as rb
        
        fp_ni = open(self.__ni_path, 'rb')

        bb = ByteBuffer(fp_ni.read(), '<')
        
        # Nodes index file format version (1)
        bb.get_short()

        # Story pack version
        version = bb.get_short()
        # Start of actual nodes list in this file (0x200 / 512)

        nodes_list = bb.get_int()
        # Size of a stage node in this file (0x2C / 44)

        node_size = bb.get_int()
        # Number of stage nodes in this file

        stage_nodes_count = bb.get_int()
        # Number of images (in RI file and rf/ folder)

        image_assets_count = bb.get_int()
        # Number of sounds (in SI file and sf/ folder)

        sound_assets_count = bb.get_int()
        # Is factory pack (boolean) set to true to avoid pack inspection by official Luniistore application

        factory_disabled = bb.get() != 0x00

        for i in range(stage_nodes_count):
            fp_ni.seek(0x200 + (i * node_size))
            bb = ByteBuffer(fp_ni.read(node_size), '<')

            ri_index = bb.get_int() # index in ri file
            si_index = bb.get_int() # index in si file

            next_node_index = bb.get_int() # index loaded from li for next in ni
            next_nodes_count = bb.get_int() # how many choices for next_node?
            # Choices Range = next_node_index -> next_node_index + next_nodes_count
            next_node_selected = bb.get_int() # huh... always 0 for some reason?

            home_node_index = bb.get_int()
            home_nodes_count = bb.get_int()
            home_node_selected = bb.get_int() # to ignore too

            controls = {
                'wheel': bb.get_short() != 0,
                'ok': bb.get_short() != 0,
                'home': bb.get_short() != 0,
                'pause': bb.get_short() != 0,
                'autoplay': bb.get_short() != 0
            }

            next_transitions = []
            if next_node_index != -1 and next_nodes_count != -1 and next_node_selected != -1:
                for count in range(next_nodes_count):
                    next_transitions.append(Transition(self.__get_abs_node_idx(next_node_index) + count))

            home_transitions = []
            if home_node_index != -1 and home_nodes_count != -1 and home_node_selected != -1:
                for count in range(home_nodes_count):
                    home_transitions.append(Transition(self.__get_abs_node_idx(home_node_index) + count))

            assets = {
                'image': None,
                'audio': None
            }

            if ri_index != -1:
                image_path = self.__rf_path + self.__ri[ri_index].replace("\\", "/")
                key = self.__lunii_generic_key
                assets['image'] = ImageAsset('image/bmp', image_path, key)

            audio = None
            if si_index != -1:
                audio_path = self.__sf_path + self.__si[si_index].replace("\\", "/")
                assets['audio'] = AudioAsset('audio/mpeg', audio_path)

            stage_node = StageNode(
                i,
                assets,
                next_transitions,
                home_transitions,
                ControlSettings(controls),
            )
            self.stage_nodes[i] = stage_node

        li_fp = Decryption(self.__lunii_generic_key, path=self.__li_path).decrypted_buffer
        li_bb = ByteBuffer(li_fp.read(), '<')

       


    def __verify_auth(self):
        # bt file is encrypted with the device key
        # It is made by ciphering the 0x40 first bytes for .ri file with device specific key.

        try:
            with open(self.__bt_path, 'rb') as fp_bt:
                bt = fp_bt.read()

            with open(self.__ri_path, 'rb') as fp_ri:
                ri = fp_ri.read(0x40)

            decryptionClient = Decryption(self.__device_key, bytes=bt)
            dec = decryptionClient.decrypted

            if dec == ri:
                self.authorized = True
            else:
                self.authorized = False
        except:
            self.authorized = False




import requests
import time
import json

url = "https://raw.githubusercontent.com/o-daneel/Lunii.RE/main/resources/packs.json"
response = requests.get(url).json()

results = []

for key in response["response"]:
    entry = response["response"][key]
    localized_infos = entry.get("localized_infos", {}).get("fr_FR", {})
    if not localized_infos:
        continue
    else:
        title = localized_infos.get("title", "")
    result = {
        "uuid": entry.get("uuid", ""),
        "title": title,
        "keywords": entry.get("keywords", ""),
        "subtitle": entry.get("subtitle", ""),
    }
    results.append(result)



json.dump(results, open("titles.json", "w"), indent=4)


# cli for lumiios
# WIP-- TODO later

# Path: lumiios/cli.py

import argparse
import logging
import os
import sys
from pathlib import Path

from . import Device, Story, utils

logger = logging.getLogger(__name__)

def main():
    parser = argparse.ArgumentParser(
        description="A Python package for working with Lunii devices, and emulating them",
        epilog="For more information, visit github.com/bastien8060/lumiios",
    )
    
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="%(prog)s 0.1.0a1",
    )

    args = parser.parse_args()

    if args.version:
        print("lumiios 0.1.0a1")
        sys.exit(0)

 


import io
import os
import json
import time
import struct
import requests
from typing import Dict, List, Optional, Union, Tuple

import vlc
import xxtea
import climage
from PIL import Image
from appdirs import AppDirs

from pkg_resources import resource_stream

class Decryption:
    def __lunii_tea_rounds(self, buffer):
        return int(1 + 52 / (len(buffer)/4))
    
    def __init__(self, key, padding=None, rounds=None, path=None, bytes=None):
        self.key = key
        self.padding = padding
        self.rounds = rounds
        self.decrypted = None
        self.decrypted_buffer = None

        
        #check if path or bytes is set
        if path is None:
            self.decrypted = xxtea.decrypt(bytes, self.key, padding=False, rounds=self.__lunii_tea_rounds(bytes))
            return

        else:
            fp = open(path, 'rb')

            # processing first block
            ciphered = fp.read(0x200)
            if len(ciphered) == 0:
                return
            
            self.decrypted = xxtea.decrypt(ciphered, self.key, padding=False, rounds=self.__lunii_tea_rounds(ciphered))

            # check if left over bytes, that are already decrypted
            left = fp.read()
            if len(left) != 0:
                self.decrypted += left

            fp.close()

            self.decrypted_buffer = io.BytesIO(self.decrypted)

            return

class ByteBuffer:
    def __init__(self, data, byte_order='!'):
        self.data = data
        self.offset = 0
        self.byte_order = byte_order

    def get(self):
        if self.offset + 1 > len(self.data):
            raise ValueError("Not enough bytes remaining in buffer")
        b = self.data[self.offset]
        self.offset += 1
        return b

    def get_short(self):
        if self.offset + 2 > len(self.data):
            raise ValueError("Not enough bytes remaining in buffer")
        s = struct.unpack_from(self.byte_order + "h", self.data, self.offset)[0]
        self.offset += 2
        return s

    def get_int(self):
        if self.offset + 4 > len(self.data):
            raise ValueError("Not enough bytes remaining in buffer")
        i = struct.unpack_from(self.byte_order + "i", self.data, self.offset)[0]
        self.offset += 4
        return i
    
    def read_n_bytes(self, n):
        if self.offset + n > len(self.data):
            raise ValueError("Not enough bytes remaining in buffer")
        b = self.data[self.offset:self.offset + n]
        self.offset += n
        return b

    def seek(self, offset, whence=0):
        if whence == 0:
            self.offset = offset
        elif whence == 1:
            self.offset += offset
        elif whence == 2:
            self.offset = len(self.data) + offset
        else:
            raise ValueError("Invalid whence argument")
        if self.offset < 0 or self.offset > len(self.data):
            raise ValueError("Invalid offset argument")

class ImageAsset:
    """
    BPM Images
    Format                                   : RLE4
    Width                                    : 320 pixels
    Height                                   : 240 pixels
    Color space                              : RGB
    Bit depth                                : 4 bits
    """
    def __init__(self, image_type: str, path: str, key: str):
        self.image_type = image_type
        self.path = path

        self.content = Decryption(key, path=path).decrypted
    
    def show(self):
        img = Image.open(io.BytesIO(self.content))
        img.save('/tmp/image.png')

        # display the image
        img = Image.open('/tmp/image.png')
    
        output = climage.convert('/tmp/image.png', 'ansi', width=40)

        print(output)
        
    def __repr__(self):
        return f"ImageAsset({self.image_type}, {self.path})"

    def __str__(self):
        return f"ImageAsset({self.image_type}, {self.path})"

class AudioAsset:
    """
    Audio
    Format                                   : MPEG Audio
    Format version                           : Version 1
    Format profile                           : Layer 3
    Duration                                 : 1 s 881 ms
    Bit rate mode                            : Variable
    Bit rate                                 : 61.8 kb/s
    Minimum bit rate                         : 32.0 kb/s
    Channel(s)                               : 1 channel
    Sampling rate                            : 44.1 kHz
    Frame rate                               : 38.281 FPS (1152 SPF)
    Compression mode                         : Lossy
    Stream size                              : 14.2 KiB (97%)
    Writing library                          : LAME3.100
    Encoding settings                        : -m m -V 4 -q 0 -lowpass 17.5 --vbr-new -b 32
    """
    def __init__(self, audio_type: str, path: str):
        self.audio_type = audio_type
        self.path = path
        
        with open(path, 'rb') as f:
            self.content = f.read()

    @property
    def is_playing(self):
        return self.player.get_state() != vlc.State.Ended

    def play(self):
        # create a VLC instance
        self.instance = vlc.Instance('--no-xlib')

        # create a new media object
        self.media = self.instance.media_new(self.path)

        # create a new media player
        self.player = self.instance.media_player_new()

        # set the media player as the media
        self.player.set_media(self.media)

        time.sleep(0.25)

        # play the media
        self.player.play()

    def release(self):
        #release the player
        self.player.release()
        self.instance.release()
        

    def __repr__(self):
        return f"AudioAsset({self.audio_type}, {self.path})"

    def __str__(self):
        return f"AudioAsset({self.audio_type}, {self.path})"
        


class Transition:
    def __init__(self, to_index: int):
        self.to_index = to_index

class ControlSettings:
    def __init__(self, dict_controls):
        self.wheel = dict_controls.get('wheel', False)
        self.ok = dict_controls.get('ok', False)
        self.home = dict_controls.get('home', False)
        self.pause = dict_controls.get('pause', False)
        self.autoplay = dict_controls.get('autoplay', False)

class StageNode:
    def __init__(self,
                 index: int,
                 assets: Tuple[ImageAsset, AudioAsset] = None,
                 next_transitions: List[Transition] = None,
                 home_transitions: List[Transition] = None,
                 control_settings: ControlSettings = None
                 ):
        self.index = index
        self.assets = assets
        self.next_transitions = next_transitions
        self.home_transitions = home_transitions
        self.control_settings = control_settings

class ActionNode:
    def __init__(self, options, metadata):
        self.options = options
        self.metadata = {
            'choice_index': metadata['choice_index'],
            'choice_count': metadata['choice_count'],
        }

class StoryMetadata:
    def __init__(self, filename=None, refresh=False):
        self.filename = filename
        if refresh:
            self.__fetch_metadata()
        self.metadata = self.load_titles()

    def __fetch_metadata(self, url: str, lang: str = "fr_FR") -> None:
        response = requests.get(url).json()

        results = []

        for key in response["response"]:
            entry = response["response"][key]
            localized_infos = entry.get("localized_infos", {}).get(lang, {})
            if not localized_infos:
                continue
            else:
                title = localized_infos.get("title", "")
            result = {
                "uuid": entry.get("uuid", ""),
                "title": title,
                "keywords": entry.get("keywords", ""),
                "subtitle": entry.get("subtitle", ""),
            }
            results.append(result)

        self.filename = self.filename or os.path.join(
            AppDirs("lunii", "lunii").user_data_dir, "titles.json"
        )
        with open(self.filename, 'w') as f:
            json.dump(results, f, indent=4)

    def load_titles(self):
        self.filename = self.filename or os.path.join(
            AppDirs("lunii", "lunii").user_data_dir, "titles.json"
        )
        if os.path.exists(self.filename):
            with open(self.filename, 'r') as f:
                return json.load(f)
        return json.load(resource_stream('lunii', 'data/titles.json'))

    def get_metadata(self, uuid):
        for story in self.metadata:
            if str(uuid).lower() in story['uuid'].lower():
                return story
        return {}



from uuid import UUID

from . import Story
from .utils import Decryption, ByteBuffer

class Device:
    def __init__(self, path):
        self.mountpoint = path
        self.devkey = None
        self.devicekey = None
        self.story_list, stories = [], []

        # external flash hardcoded value
        # 91BD7A0A A75440A9 BBD49D6C E0DCC0E3
        self.raw_key_generic = [0x91BD7A0A, 0xA75440A9, 0xBBD49D6C, 0xE0DCC0E3]
        self.lunii_generic_key = self.__vectkey_to_bytes(self.raw_key_generic)

        self.__get_hw_info()
        self.__get_stories()

    def populate(self):
        self.__get_stories()

    def parse_stories(self):
        self.__parse_stories()

    def parse_story(self, uuid):
        if uuid in self.story_list:
            return Story(self.mountpoint, uuid, (self.device_key, self.lunii_generic_key))
        return None
        
    def __get_hw_info(self):
        md_path = f"{self.mountpoint}/.md"

        with open(md_path, 'rb') as fp_md:
            fp_md.seek(fp_md.tell() + 6)

            self.fw_vers_major = int.from_bytes(fp_md.read(2), 'little')
            self.fw_vers_minor = int.from_bytes(fp_md.read(2), 'little')
            self.snu = int.from_bytes(fp_md.read(8), 'little')
            
            fp_md.seek(0x100)
            self.raw_devkey = fp_md.read(0x100)

            # I'll rewrite that with utils.py:ByteBuffer

            decryptionClient = Decryption(self.lunii_generic_key, bytes=self.raw_devkey)
            dec = decryptionClient.decrypted

            # Reordering Key components
            self.device_key = dec[8:16] + dec[0:8]

    def __vectkey_to_bytes(self, key_vect):
        joined = [k.to_bytes(4, 'little') for k in key_vect]
        return b''.join(joined)
    
    def __get_stories(self):
        self.story_list = []
        with open(self.mountpoint + '/.pi', 'rb') as fp_pi:
            loop_again = True
            while loop_again:
                next_uuid = fp_pi.read(16)
                if next_uuid:
                    self.story_list.append(UUID(bytes=next_uuid))
                else:
                    loop_again = False

    def __parse_stories(self):
        self.stories = []
        for story in self.story_list:
            self.stories.append(Story(self.mountpoint, story, (self.device_key, self.lunii_generic_key)))
      


from setuptools import setup, find_packages

setup(
    name='lunii',
    version='0.1.0a2',
    description='A Python package for working with Lunii devices, and emulating them',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Bastien Saidi',
    author_email='saidibastien@gmail.com',
    url='https://github.com/bastien8060/lumiios',
    packages=['lunii'],
    package_data={'lunii': ['data/titles.json']},
    install_requires=[
        # List the required packages here
        'xxtea >= 2.0.0',
        'climage >= 0.1.2',
        'Pillow >= 7.2.0',
        'python-vlc >= 2.2.6100',
        'appdirs',
    ],
    entry_points={
        'console_scripts': ['lunii-cli=lumiios.cli:main']
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Education',
        'Topic :: Multimedia',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries',
        'Topic :: System :: Emulators'
    ],
)